# coding: utf8
"""Utilities for tankobon.
"""

import random
import re
from typing import Union, Optional

import bs4
import requests
import requests_random_user_agent

Number = Union[int, float]

# Downloader config
BSOUP_PARSER = "html5lib"  # if you want, change to lxml for faster parsing
TIMEOUT = 5
COOLDOWN = 0.5
THREADS = 8

# map to mimetypes (mimetypes module sucks)
FILE_EXTENSIONS = {"image/jpeg": "jpg", "image/png": "png", "image/gif": "gif"}


def get_file_extension(response: requests.models.Response) -> str:
    """Get a extension from a response by parsing its mimetype.

    Args:
        response: The response object (from requests.get, et al.)
    """

    content_type = response.headers["Content-Type"]
    return f".{FILE_EXTENSIONS[content_type]}"


def get_soup(
    *args, encoding: Optional[str] = None, parser: str = BSOUP_PARSER, **kwargs
) -> bs4.BeautifulSoup:
    """Get a url as a BeautifulSoup.

    Args:
        *args: See get_url.
        encoding: The encoding to decode.
            Defaults to the autodetected encoding (by requests).
        parser: The parser to use.
            Must be 'html.parser', 'html5lib' or 'lxml'.
        **kwargs: See get_url.
    """

    response = requests.get(*args, **kwargs)
    if encoding is not None:
        html_text = response.content.decode(encoding)
    else:
        html_text = response.text

    return bs4.BeautifulSoup(html_text, parser)



# coding: utf8
"""tankobon (漫画): Manga downloader and scraper."""

import abc
import json
import logging
import multiprocessing as mproc
import pathlib
from collections.abc import MutableMapping
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

import bs4
import requests

from . import utils
from .exceptions import CacheError

Chapter = Tuple[str, str, str]

_log = logging.getLogger("tankobon")


class Cache(MutableMapping):
    """A cache/downloader for Manga pages, stored on disk.

    Args:
        path: The path to store the cache at.
        database: The existing manga database.
            If not specified, the database will be loaded from 'index.json' in path.
    """

    CACHE_FILENAME = "index.json"

    def __init__(
        self, path: Union[str, pathlib.Path], database: Optional[dict] = None
    ) -> None:
        self.path = pathlib.Path(path)
        self.cachepath = self.path / self.CACHE_FILENAME

        # create path if it does not exist
        self.path.mkdir(exist_ok=True)

        if database is None:
            self.reload_cache()
        else:
            self.database = database

    def __getitem__(self, key):
        return self.database[key]

    def __setitem__(self, key, value):
        self.database[key] = value

    def __delitem__(self, key):
        del self.database[key]

    def __iter__(self):
        return iter(self.database)

    def __len__(self):
        return len(self.database)

    def __repr__(self):
        return f"{type(self).__name__}({self.database})"

    def reload_cache(self) -> None:
        """Load the existing cache database from self.path into self.database.

        Raises:
            CacheError, if the database does not exist.
        """

        try:
            cache = self.path / self.CACHE_FILENAME
            with cache.open() as f:
                self.database = json.load(f)
            _log.debug("reloaded cache")

        except FileNotFoundError:
            raise CacheError(f"cache at path {str(cache)} does not exist")

    def download_chapters(
        self,
        ids: Optional[List[str]] = None,
        force: bool = False,
        threads: int = utils.THREADS,
    ) -> None:
        """Download chapters, caching its pages on disk.
        Ignores any existing chapter data on disk (downloads anyway).

        Args:
            ids: The page ids. Defaults to all chapters.
            force: Whether or not to re-download chapters, regardless if they are
                already downloaded. Defaults to False.
            threads: The number of threads to use to download the pages.
                Defaults to utils.THREADS (8).
        """

        if ids is None:
            ids = self.database["chapters"].keys()

        for id in ids:

            chapter_path = self.path / id
            if chapter_path.exists() and not force:
                _log.info(f"skipping chapter {id}")
                continue
            _log.info(f"downloading chapter {id}")

            chapter_path.mkdir(exist_ok=True)
            urls = self.database["chapters"][id]["pages"]

            with ThreadPool(threads) as pool:
                responses = pool.imap(requests.get, urls)

                for page_number, response in enumerate(responses):
                    _log.debug("downloading page %s", page_number)

                    page_path = (
                        chapter_path
                        / f"{page_number}{utils.get_file_extension(response)}"
                    )

                    with page_path.open(mode="wb+") as f:
                        f.write(response.content)

    def close(self):
        with self.cachepath.open("w+") as f:
            json.dump(self.database, f)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if tb is None:
            self.close()


class GenericManga(abc.ABC):
    """A generic manga website.

    Args:
        database: The inital database of the manga as a dictionary.
            It must be in this format:
            {
                "title": "...",  # manga title/name
                "url": "...",  # manga index (chapter listing)
                "chapters": {...}  # cached chapter info, automatically generated
            }
            where 'title' and 'url' must be specified, and 'chapters' may be an empty
            dict.
        update: Whether or not to download and parse the index, adding any new
            chapters. Defaults to True.

    Attributes:
        database (dict): see args
        soup (bs4.BeautifulSoup): The soup of the HTML.
    """

    # you should overrride this
    DEFAULTS: Dict[str, Any] = {}

    def __init__(
        self,
        database: Optional[Union[dict, Cache]] = None,
        update: bool = True,
    ) -> None:
        if database is None:
            self.database = self.DEFAULTS
        else:
            self.database = database  # type: ignore

        self.soup = utils.get_soup(self.database["url"])
        if update:
            self.refresh()

    def __getattr__(self, key):
        value = self.database.get(key)
        if value is None:
            raise AttributeError
        return value

    def query(self, query: str) -> Any:
        """A much nicer way to get values from a dictionary.

        Args:
            query: The query, in the format 'key1:keyX' where ':' seperates
                successive keys and is evaluated to ['key1']['keyX'].
                Keys must not have ':' in their names.
        """

        value = None

        if not self.database:
            raise ValueError("database is empty")

        for key in query.split(":"):
            value = value or self.database
            value = value[key]
        return value

    @staticmethod
    def is_link(tag: bs4.element.Tag) -> bool:
        """Check whether a BeautifulSoup tag is a link.

        Args:
            tag: The tag.

        Returns:
            True if so, otherwise False.
        """

        return (tag.name == "a") and (tag.get("href") is not None)

    def is_parsed(self, id: str) -> bool:
        """Check whether a chapter has already been parsed to get its page URLs.

        Args:
            id: The chapter id.

        Returns:
            True if so, otherwise False.
        """

        try:
            return bool(self.database["chapters"][id]["url"])
        except KeyError:
            return False

    @abc.abstractmethod
    def parse_chapters(self) -> Generator[Chapter, None, None]:
        """Parse all chapters from the soup.

        Yields:
            A three-tuple of (chapter_id, chapter_title, chapter_url).
        """

        raise NotImplementedError

    @abc.abstractmethod
    def parse_pages(self, id: str, force: bool = False) -> list:
        """Parse all pages from a chapter.
        The chapter's info must have already been cached into the database.

        The chapter webpage's HTML should be downloaded and parsed for page links.
        The page URLs will be cached in the database and returned if requested
        subsequently (won't parse again).

        Args:
            id: The chapter id.
            force: Whether or not to re-parse the chapter webpage for pages,
                regardless of whether or not the pages have already been parsed.
                Defaults to False.

        Returns:
            A list of page URLs.
        """

        raise NotImplementedError

    def refresh(self, force: bool = False) -> None:
        """Refresh the database, adding any new chapter info.
        Does not download the chapter webpages (under the 'pages' key).

        Args:
            force: Whether or not to overwrite any existing chapters with newer data.
        """

        for id, title, url in self.parse_chapters():
            if self.is_parsed(id) and not force:
                continue
            self.database["chapters"][id] = {"url": url, "title": title, "pages": []}

    @property
    def existing_chapters(self) -> Generator[Chapter, None, None]:
        """Generate existing chapters cached into the database.

        Yields:
            A three-tuple of (id, title, url).
        """
        for id, chapter in self.database["chapters"].items():
            yield id, chapter["title"], chapter["url"]

    def parse_all(self, threads: int = utils.THREADS, force: bool = False) -> dict:
        """Parse all chapters, adding their page URLs to their info.

        Args:
            threads: How many threads to use to speed up parsing.
                Defaults to THREADS (8).
            force: Whether or not to parse all chapters, regardless of whether or not
                the chapter's pages have already been parsed. Defaults to False.

        Returns:
            The info of all chapters mapped to their ids.
        """

        def _parse_all(args):
            id, title, url = args
            if self.database["chapters"][id]["pages"] and not force:
                _log.info(f"skipping {id}")
                return
            pages = self.parse_pages(id, force=True)
            _log.info(f"parsed {id}")
            return id, {"title": title, "url": url, "pages": pages}

        with ThreadPool(threads) as pool:
            results = pool.imap_unordered(_parse_all, self.existing_chapters)

            for result in results:
                if result is None:
                    continue
                id, chapter = result
                self.database[id] = chapter

        return self.database



# coding: utf8
"""Bootstrap for Komi Can't Communicate."""

import json
import pathlib
import re
from typing import Generator

from tankobon.base import GenericManga, Chapter
from tankobon.utils import get_soup


class Manga(GenericManga):

    IMGHOST = "blogspot.com"
    RE_TITLE = re.compile(r"(?:.*)Chapter (\d+(?:\.\d)?) *[:\-] *(.+)\Z")
    # not used, kept here for reference
    RE_URL = re.compile(r"(?:.*)?chapter-(\d+(?:-\d)?)-([\w\-]*)/?\Z")

    DEFAULTS = {
        "title": "Komi Can't Communicate",
        "url": "https://komi-san.com",
        "chapters": {},
    }

    def page_is_valid(self, url: str) -> bool:
        """Check if a page url is valid.

        Returns:
            True if so, otherwise False.
        """

        return self.IMGHOST in url

    def parse_chapters(self) -> Generator[Chapter, None, None]:
        """Generate all chapters in the manga.

        Yields:
            A three-tuple of (id, title, url) for each chapter.
        """

        # get rid of section
        section = self.soup.find("section", class_="widget ceo_latest_comics_widget")
        if section is not None:
            section.decompose()

        for tag in self.soup.find_all("a"):

            if not self.is_link(tag):
                continue
            href = tag.get("href")
            title = tag.text

            match = self.RE_TITLE.findall(title)

            if match:
                id, title = match[0]
                yield id, title, href

    def parse_pages(self, id: str, force: bool = False) -> list:
        pages = []
        soup = get_soup(self.database["chapters"][id]["url"], encoding="utf-8")

        if not (self.is_parsed(id) and not force):

            for link in soup.find_all("img"):
                src = link["src"]
                if self.page_is_valid(src):
                    pages.append(src)

            self.database["chapters"][id]["pages"] = pages

        return self.database["chapters"][id]["pages"]


# coding: utf8

import json
import logging
import pathlib
import zipfile
from functools import partial

import click

from tankobon.__version__ import __version__
from tankobon.base import Cache
from tankobon.bootstraps import Bootstrap, update_index
from tankobon.utils import THREADS

# monkey-patch options
click.option = partial(click.option, show_default=True)  # type: ignore
VERBOSITY = (
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
)
_log = logging.getLogger("tankobon")

CACHEPATH = pathlib.Path.home() / "Documents" / "tankobon"
BOOTSTRAP_PATH = pathlib.Path(__file__).parent / "bootstraps"


@click.group()
@click.version_option(__version__)
@click.option(
    "-v", "--verbose", "verbosity", help="be more chatty", default=4, count=True
)
def cli(verbosity):
    """Manga browser/downloader."""
    # set up logger
    _log.setLevel(VERBOSITY[verbosity - 1])


@cli.command()
def list():
    """List all mangas available."""
    for manga in Bootstrap.available:
        name = Bootstrap(manga).manga.DEFAULTS["title"]
        click.echo(f"{manga} ({name})")


@cli.command()
@click.argument("name")
def info(name):
    """Get info on a manga bootstrap."""
    manga = Bootstrap(name)()
    click.echo(f"Title: {manga.title}")
    click.echo(f"URL: {manga.url}")
    click.echo(f"Chapters:")
    for chapter in sorted(manga.chapters, key=float):
        chapter_info = manga.chapters[chapter]
        click.echo(f"{chapter}: {chapter_info['title']} ({chapter_info['url']})")


@cli.command()
@click.argument("name")
@click.option("-p", "--path", help="where to download to", default=".")
@click.option(
    "-t",
    "--threads",
    help="how many threads to use to download the manga",
    default=THREADS,
)
@click.option(
    "-r", "--refresh", help="whether or not to parse any new chapters", is_flag=True
)
@click.option(
    "-c",
    "--chapters",
    help="which chapters to download, seperated by slashes",
    default="all",
)
def download(name, path, threads, refresh, chapters):
    """Download a manga with name."""
    manga = Bootstrap(name)()

    if refresh:
        manga.refresh()

    with Cache(path, database=manga.database) as c:
        if chapters == "all":
            chapters = None
        else:
            chapters = chapters.split("/")

        c.download_chapters(ids=chapters, threads=threads)


@cli.command()
def update():
    """Update all manga bootstraps (in the index)."""
    update_index()


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['tankobon', 'tankobon.bootstraps']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.9.1',
 'click>=7.1.2',
 'coloredlogs>=14.0',
 'html5lib>=1.1',
 'requests>=2.24.0',
 'requests-random-user-agent>=2020.10.19']

extras_require = \
{'gui': ['pyglet>=1.5.8'], 'jsonschema': ['jsonschema>=3.2.0']}

entry_points = \
{'console_scripts': ['tankobon = tankobon.cli:cli']}

setup(name='tankobon',
      version='2.0.0a3',
      description='Manga browser/downloader.',
      author='Ong Yong Xin',
      author_email='ongyongxin2020+github@gmail.com',
      url='https://github.com/ongyx/tankobon',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



# coding: utf8
"""Manga browser/downloader."""

import logging

import coloredlogs

from .__version__ import __version__  # noqa: f401

coloredlogs.install(
    fmt=" %(levelname)-8s :: %(message)s",
    logger=logging.getLogger("tankobon"),
)



# coding: utf8
"""Bootstraps are scripts used to scrape manga off a webpage.

A manga bootstrap consists of a Python module with a Manga class in this directory
(name.py), corresponding to a JSON file (name.json) in a compressed zipfile
(INDEX.zip) to save space.

`name` is the manga ID, i.e `Manga-Name: with Punctuation!`-> `manga_name_with_punctuation`.
The name must be a valid Python identifier (recommended to use `snake_case`).
(Because it will be imported as a module.)

Example (`name.py`):

```python
from tankobon.base import GenericManga

# Every bootstrap must have a 'Manga' class which inherits from tankobon.base.GenericManga.
# The 'parse_chapters' and 'parse_pages' methods of GenericManga must be overriden.

class Manga(GenericManga):
    def parse_chapters(self):
        ...
    def parse_pages(self, ...):
        ...
```

The `parse_all` method of the Manga class must emit a manifest conforming to the
Manga schema (see tankobon.schema.SCHEMA). It has already been defined in GenericManga,
so most of the time there is no need to override it. It *just* works.
(See documentation for both functions in tankobon.base.GenericManga on how to correctly implement them.)

The emitted manifest should be added to `INDEX.zip` in this directory. If
recompressing, use DEFLATE.
(I should problably add a way to automate this...)

To load an existing bootstrap:

```python
from tankobon.bootstraps import Bootstrap
# load the associated Manga object from the bootstrap.
manga = Bootstrap("manga_name")
# load the existing manifest for the manga (if avaliable).
manifest = bootstraps.retrieve_manifest("manga_name")
```

"""

import importlib
import io
import json
import logging
import pathlib
import zipfile

from tankobon.base import GenericManga

_log = logging.getLogger("tankobon")

BOOTSTRAP_PATH = pathlib.Path(__file__).parent
INDEX = BOOTSTRAP_PATH / "INDEX.zip"


class Bootstrap(object):
    """Helper to load Manga classes from bootstraps.

    Usage:

    manga_bootstrap = Bootstrap('manga_name')
    # the raw Manga object
    Manga = manga_bootstrap.manga
    # preinitalised with manifest
    manga = manga_bootstrap(
        ...  # other args
    )

    Args:
        name: The bootstrap name.

    Attributes:
        name (str): Bootstrap name.
        manga (tankobon.base.GenericManga): The uninitalised Manga class (if you want to subclass).
        available (list): All loadable bootstraps.
    """

    available = set()

    for pyfile in BOOTSTRAP_PATH.glob("*.py"):
        if not pyfile.stem == "__init__":
            available.add(pyfile.stem)

    def __init__(self, name: str, index_path: pathlib.Path = INDEX) -> None:
        if name not in self.available:
            raise ValueError(f"bootstrap '{name}' does not exist")

        try:
            self._bootstrap_module = importlib.import_module(
                f"tankobon.bootstraps.{name}"
            )
        except ModuleNotFoundError as err:
            raise ValueError(f"failed loading bootstrap '{name}': {err}")

        self.name = name
        self.index_path = index_path
        self.pyfile = BOOTSTRAP_PATH / f"{name}.py"
        _log.debug("initalised bootstrap for %s", name)

    @property
    def Manga(self) -> type:
        # mypy dosen't like dynamic imports
        return self._bootstrap_module.Manga  # type: ignore

    def __call__(self, *args, **kwargs) -> GenericManga:
        """Initalise the Manga object using the database loaded from the bootstrap.

        Args:
            *args: Passed to the Manga constructor.
            **kwargs: Passed to the Manga constructor.

        Returns:
            The initalised Manga object.
        """

        with zipfile.ZipFile(str(self.index_path)) as zf:
            try:
                with zf.open(f"{self.name}.json") as f:
                    manifest = json.load(f)

            except (KeyError, zipfile.BadZipFile):
                # no manifest available, or zipfile is corrupted
                manifest = None

        _log.debug("loaded bootstrap for %s", self.name)
        if manifest is not None:
            return self.Manga(manifest, *args, **kwargs)

        return self.Manga(*args, **kwargs)


def update_index(index_path: pathlib.Path = INDEX) -> None:
    """Update INDEX.zip with the newest bootstrap files.

    Args:
        index_path: The path to the index file. Defaults to INDEX.
    """
    # use a seperate buffer, because zipfile will complain about duplicate names...
    buffer = io.BytesIO()
    zf_buffer = zipfile.ZipFile(buffer, mode="x")

    with zipfile.ZipFile(str(index_path)) as zf:
        for name in Bootstrap.available:
            _log.info("updating index for manga %s", name)
            filename = f"{name}.json"
            manga = Bootstrap(name).Manga(database=json.loads(zf.read(filename)))
            manga.parse_all()
            zf_buffer.writestr(filename, json.dumps(manga.database))

    zf_buffer.close()

    with index_path.open(mode="wb") as f:
        f.write(buffer.getvalue())



# coding: utf8
"""Schema validifier for tankobon manifests.
"""

try:
    import jsonschema
except ImportError:
    jsonschema = None

# Schema for manga metadata.
# It should be emitted in this format as JSON.
SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Manga",
    "description": "A manga/comic, hosted on a webserver.",
    "type": "object",
    "definitions": {
        "pages": {"type": "array", "items": {"type": "string"}},
        "chapter": {
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "url": {"type": "string"},
                "pages": {"$ref": "#/definitions/pages"},
            },
        },
    },
    "properties": {
        "title": {"type": "string"},
        "url": {"type": "string"},
        "chapters": {
            "type": "object",
            "propertyNames": {"pattern": "^([0-9]+(\\.[0-9])?)$"},
            "$ref": "#/definitions/chapter",
        },
    },
    "required": ["title", "url", "chapters"],
}